/** ZCT/search.c--Created 070905 **/

/** Copyright 2008 Zach Wegner **/
/*
 * This file is part of ZCT.
 *
 * ZCT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ZCT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ZCT.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "probe.h"
#include "smp.h"

#define SEARCH_CALL(d, p, a, b, fm, nt, ss) \
	do { \
		search_call((sb), FALSE, (d), (p), (a), (b), (fm), (nt), (ss)); \
		sb++; \
		goto end; \
	} while (FALSE); case ss:

#define QSEARCH_CALL(d, p, a, b, fm, nt, ss) \
	do { \
		search_call((sb), TRUE, (d), (p), (a), (b), (fm), (nt), (ss)); \
		sb++; \
		goto end; \
	} while (FALSE); case ss:

#define RETURN(r) \
	do { \
		sb->return_value = r; \
		sb--; \
		goto end; \
	} while (FALSE)

/**
search():
Do an iterative alpha-beta search.
Created 070905; last modified 072108
**/
VALUE search(SEARCH_BLOCK *sb)
{
	VALUE r;
#ifdef SMP
	int x;
	int process;
	int last_help_time = -1;
	int help_count = 0;
	int done[4] = {0,0,0,0};
	SEARCH_BLOCK *temp;
#endif

	while (TRUE)
	{
#ifdef DEBUG_SEARCH
		verify();
		debug(sb);
#endif
#ifdef SMP
		if (*board.split_ply != -1)
			ASSERT(sb->ply >= *board.split_ply);
#endif
		switch (sb->search_state)
		{
/**
The main search.
**/
case SEARCH_START:
#ifdef SMP
		smp_block[board.id].nodes++;
#else
		zct->nodes++;
#endif
		if (sb->ply >= zct->max_depth_reached)
			zct->max_depth_reached = sb->ply;
		if (sb->ply >= MAX_PLY - 1)
			RETURN(evaluate(&board.eval_stack[sb->ply]));
		if (board.fifty_count >= 100 && !is_mate())
			RETURN(DRAW);
		if (is_repetition(sb->ply > 1 ? 2 : 3))
			RETURN(DRAW);

		/* mate-distance pruning */
		sb->alpha = MAX(sb->alpha, -MATE + sb->ply);
		sb->beta = MIN(sb->beta, MATE - sb->ply - 1);
		if (sb->alpha >= sb->beta)
			RETURN(sb->alpha);

		/* material draw pruning */
		if (!can_mate(board.side_tm))
		{
			sb->beta = MIN(sb->beta, DRAW);
			if (sb->alpha >= sb->beta)
				RETURN(sb->alpha);
		}
		if (!can_mate(board.side_ntm))
		{
			sb->alpha = MAX(sb->alpha, DRAW);
			if (sb->alpha >= sb->beta)
				RETURN(sb->alpha);
		}
		/* Probe the hash table. */
		if (hash_probe(sb, FALSE))
			RETURN(sb->alpha);

#ifdef EGBB
		/* Probe the bitbases. */
		if (egbb_is_loaded && pop_count(board.occupied_bb) <= 4)
//		&& (board.game_entry->capture != EMPTY ||
//			board.piece[MOVE_TO((sb-1)->move)] == PAWN))
		{
			if (probe_bitbases(&r))
			{
				if (r >= WIN_SCORE) /* prefer wins closer to root */
					r -= sb->ply * WIN_PLY;
				else if (r <= -WIN_SCORE)
					r += sb->ply * WIN_PLY;
				hash_store(sb, NULL_MOVE, r, HASH_LOWER_BOUND, FALSE);
				RETURN(r);
			}
		}
#endif

		sb->check = check_squares();
		/* Nullmove search. */
		if (should_nullmove(sb))
		{
			sb->move_made = TRUE;
			make_move(NULL_MOVE);
			SEARCH_CALL(sb->depth - null_reduction(sb) - PLY, sb->ply + 1,
				-sb->beta, -sb->beta + 1, sb->last_move, NODE_ALL,
				SEARCH_NULL_1)
			r = -(sb + 1)->return_value;
			unmake_move();
			sb->move_made = FALSE;

			if (r >= sb->beta)
			{
				hash_store(sb, NULL_MOVE, r, HASH_LOWER_BOUND, FALSE);
				RETURN(r);
			}
			/* If the null move failed low, we can use the refutation for move
				ordering on the next ply. */
			else
			{
				board.threat_move[sb->ply + 1] = MOVE_COMPARE(board.pv_stack
					[sb->ply + 1][0]);
				if (moves_are_connected((sb - 1)->move,
					board.threat_move[sb->ply + 1]))
				{
					/* If we reduced on the last ply, check if the move was
						actually a threat that allowed the refutation. */
					if ((sb - 1)->extension < 0)
						RETURN(sb->alpha);
				}
			}
			/* Mate threat. The side that null moved was mated immediately. */
			if (r == -MATE + sb->ply + 2)
				sb->threat = TRUE;
		}

		/* Internal iterative deepening */
		if (sb->hash_move == NULL_MOVE && sb->depth >= 2 * PLY &&
			(sb->node_type == NODE_PV || sb->node_type == NODE_CUT))
		{
			SEARCH_CALL(sb->depth - 2 * PLY, sb->ply, sb->alpha, sb->beta,
				sb->last_move, sb->node_type, SEARCH_IID)
			r = (sb + 1)->return_value;
			if (r > sb->alpha)
				sb->hash_move = board.pv_stack[sb->ply][0];
		}

		/* The main move loop. */
		while (TRUE)
		{
#ifdef SMP
case SEARCH_JOIN:
			if (*board.split_ply == sb->ply)
			{
				sb->move = smp_select_move();
				SMP_DEBUG(print("cpu %i at %i got move %M, score %i\n",
					board.id, (*board.split_point)->id, sb->move,
					MOVE_SCORE(sb->move)));
			}
			else
#endif
				sb->move = select_move(sb);

			if (!sb->move)
				break;

			/* futility pruning */
			if (sb->moves > 0 && move_is_futile(sb, sb->move))
				continue;
			if (!make_move(sb->move))
				continue;
			sb->moves++;
			sb->move_made = TRUE;

			/* extensions */
			sb->extension = extend(sb);

			/* late move reductions */
			if (should_reduce(sb, sb->move))
				sb->extension = -PLY;

			/* Moves after the first in PV nodes are searched with null windows.
				This is known as Principal Variation Searching. */
			if (sb->moves > 1 && sb->node_type == NODE_PV)
			{
				SEARCH_CALL(sb->depth + sb->extension - PLY, sb->ply + 1,
					-sb->alpha - 1, -sb->alpha, sb->last_move, NODE_CUT,
					SEARCH_1)
				r = -(sb + 1)->return_value;
				if (r > sb->alpha)
				{
					/* Reset the extension if this move has been reduced. */
					sb->extension = MAX(sb->extension, 0);
					SEARCH_CALL(sb->depth + sb->extension - PLY, sb->ply + 1,
						-sb->beta, -sb->alpha, sb->last_move, NODE_PV, SEARCH_2)
					r = -(sb + 1)->return_value;
				}
			}
			else
			{
				SEARCH_CALL(sb->depth + sb->extension - PLY, sb->ply + 1,
					-sb->beta, -sb->alpha, sb->last_move,
					NODE_FLIP(sb->node_type), SEARCH_3)
				r = -(sb + 1)->return_value;
				/* reduced move research */
				if (r > sb->alpha && sb->extension < 0)
				{
					sb->extension = 0;
					SEARCH_CALL(sb->depth + sb->extension - PLY, sb->ply + 1,
						-sb->beta, -sb->alpha, sb->last_move,
						NODE_FLIP(sb->node_type), SEARCH_4)
					r = -(sb + 1)->return_value;
				}
			}
			unmake_move();
			sb->move_made = FALSE;

			/* Check the score against the alpha+beta bounds. */
			if (r > sb->best_score)
				sb->best_score = r;
			if (r > sb->alpha)
			{
				sb->alpha = r;
				sb->pv_found = sb->moves;
				board.pv_stack[sb->ply][0] = sb->move;
				copy_pv(board.pv_stack[sb->ply] + 1,
					board.pv_stack[sb->ply + 1]);
#ifdef SMP
				/* Update the bounds in the split point for our siblings. */
				if (*board.split_ply == sb->ply)
				{
					merge(sb);
					/* When we get a cutoff in a split point, then there is no
						work left, so we must wait for another split point. */
					if (sb->search_state == SEARCH_WAIT)
						goto end;
				}
#endif
				/* Fail high. */
				if (r >= sb->beta)
				{
					heuristic(sb->move, sb);
					zct->fail_high_nodes++;
					if (sb->moves == 1)
						zct->fail_high_first++;
					hash_store(sb, sb->move, r, HASH_LOWER_BOUND, FALSE);
					RETURN(r);
				}
			}
		}
#ifdef SMP
		/* If this is a split ply, we have to detach from the split point. */
		if (*board.split_ply == sb->ply)
		{
			detach(sb);
			SMP_DEBUG(print("cpu %i detached %E\n", board.id, sb));
			if (sb->search_state == SEARCH_WAIT)
				goto end;
		}
#endif
		/* No moves, check or stale mate. */
		if (sb->moves == 0)
		{
			if (sb->check)
				r = -MATE + sb->ply;
			else
				r = DRAW;
			/* Is this worth it? */
			hash_store(sb, NULL_MOVE, r, HASH_EXACT_BOUND, FALSE);
			RETURN(r);
		}
		/* Store in the hash table, plus update some statistics/heuristics. */
		if (sb->pv_found)
		{
			zct->pv_nodes++;
			if (sb->pv_found == 1)
				zct->pv_first++;
			heuristic_pv(board.pv_stack[sb->ply][0], sb);
			hash_store(sb, board.pv_stack[sb->ply][0], sb->alpha,
				HASH_EXACT_BOUND, FALSE);
		}
		else
		{
			zct->fail_low_nodes++;
			hash_store(sb, NULL_MOVE, sb->best_score, HASH_UPPER_BOUND, FALSE);
		}
		RETURN(sb->best_score);
/**
The quiescence search.
**/
case QSEARCH_START:
#ifdef SMP
		smp_block[board.id].q_nodes++;
#else
		zct->q_nodes++;
#endif
		if (sb->ply >= zct->max_depth_reached)
			zct->max_depth_reached = sb->ply;
		if (sb->ply >= MAX_PLY - 1)
			RETURN(evaluate(&board.eval_stack[sb->ply]));
		if (board.fifty_count >= 100 && !is_mate())
			RETURN(DRAW);
		if (is_repetition(sb->ply > 1 ? 2 : 3))
			RETURN(DRAW);

		/* mate-distance pruning */
		sb->alpha = MAX(sb->alpha, -MATE + sb->ply);
		sb->beta = MIN(sb->beta, MATE - sb->ply - 1);
		if (sb->alpha >= sb->beta)
			RETURN(sb->alpha);

		/* material draw pruning */
		if (!can_mate(board.side_tm))
		{
			sb->beta = MIN(sb->beta, DRAW);
			if (sb->alpha >= sb->beta)
				RETURN(sb->alpha);
		}
		if (!can_mate(board.side_ntm))
		{
			sb->alpha = MAX(sb->alpha, DRAW);
			if (sb->alpha >= sb->beta)
				RETURN(sb->alpha);
		}

		/* Probe the hash table. */
		if (hash_probe(sb, TRUE))
			RETURN(sb->alpha);
#ifdef EGBB
		/* Probe the bitbases. */
		if (egbb_is_loaded && pop_count(board.occupied_bb) <= 5)
		{
			if (probe_bitbases(&r))
			{
				sb->depth = 1;
				hash_store(sb, NULL_MOVE, r, HASH_LOWER_BOUND, FALSE);
				hash_store(sb, NULL_MOVE, r, HASH_LOWER_BOUND, TRUE);
				RETURN(r);
			}
		}
#endif

		sb->check = check_squares();
		/* If we are in check and depth is sufficient, try all evasions. */
		if (sb->check && sb->depth > 0)
		{
			sb->threat = TRUE;
			sb->select_state = SELECT_HASH_MOVE;
			if (sb->last_move - sb->first_move <= 1)
				sb->depth++;
//			if (sb->last_move - sb->first_move == 1)
//				sb->depth += 2;
		}
		/* Normal quiescence stand pat score, plus move generation. */
		else
		{
			/* Evaluate for the stand pat. */
			r = sb->best_score = evaluate(&board.eval_stack[sb->ply]);
			if (r > sb->alpha)
			{
				sb->alpha = r;
				if (r >= sb->beta)
				{
					/* Is this worth it? */
					hash_store(sb, NULL_MOVE, r, HASH_LOWER_BOUND, TRUE);
					RETURN(r);
				}
			}
			sb->select_state = SELECT_GEN_CAPS;
		}

moves:	/* Main quiescence move loop */
		while ((sb->move = select_move(sb)) != NULL_MOVE)
		{
			/* Make the move while checking legality. */
			if (!make_move(sb->move))
				continue;
			sb->moves++;
			sb->move_made = TRUE;
			/* Search the next ply. */
			QSEARCH_CALL(sb->depth - 1, sb->ply + 1, -sb->beta, -sb->alpha,
				sb->last_move, NODE_FLIP(sb->node_type), QSEARCH_1)
			r = -(sb + 1)->return_value;
			unmake_move();
			sb->move_made = FALSE;

			if (r > sb->best_score)
				sb->best_score = r;
			if (r > sb->alpha)
			{
				sb->alpha = r;
				sb->pv_found = sb->moves;
				board.pv_stack[sb->ply][0] = sb->move;
				copy_pv(board.pv_stack[sb->ply] + 1,
					board.pv_stack[sb->ply + 1]);

				if (r >= sb->beta)
				{
					heuristic(sb->move, sb);
					hash_store(sb, sb->move, r, HASH_LOWER_BOUND, TRUE);
					RETURN(r);
				}
			}
		}
		/* If all captures were unsuccessful, and the depth is sufficient, try
			some checking moves. */
		if (!sb->pv_found && !sb->threat && sb->depth > 0)
		{
			sb->select_state = SELECT_GEN_CHECKS;
			sb->threat = TRUE;
			goto moves;
		}
		/* We can detect checkmate if all quiescence nodes up to now have
			included check evasions. */
		if (sb->moves == 0 && sb->check && sb->depth > 0)
		{
			hash_store(sb, NULL_MOVE, -MATE + sb->ply, HASH_EXACT_BOUND, TRUE);
			RETURN(-MATE + sb->ply);
		}
		/* Store the result in the hash table and return. */
		if (sb->pv_found)
			hash_store(sb, board.pv_stack[sb->ply][0], sb->best_score,
				HASH_EXACT_BOUND, TRUE);
		else
			hash_store(sb, NULL_MOVE, sb->alpha, HASH_UPPER_BOUND, TRUE);
		RETURN(sb->best_score);
#ifdef SMP
case SEARCH_WAIT:
		/* If a split point already exists, attach to that. */
		if (smp_data->split_count > 0)
		{
			for (x = 0; x < MAX_SPLIT_POINTS; x++)
				if (split_point[x].active && !split_point[x].no_moves_left &&
					split_point[x].sb->last_move -
					split_point[x].sb->first_move > 3)
					break;
			if (x < MAX_SPLIT_POINTS)
			{
				SMP_DEBUG(print("cpu %i attaching to known split point %i\n",
					board.id, x));
				goto found_split_point;
			}
		}
		/* last_help_time is zero or more, meaning we have sent a HELP message
			and we're waiting for a search state from an active processor. */
		if (last_help_time != -1)
		{
			/* Increment the counter, indicating how long we've been waiting. */
			last_help_time++;
			x = -1;
			for (process = 0; process < zct->process_count; process++)
			{
				if (get_tree_block(process, board.id)->used)
				{
					SMP_DEBUG(print("cpu %i looking at split points...\n",
						board.id));
					x = find_split_point(process);
					help_count--;
					if (x != -1)
						break;
				}
			}
			/* We didn't find any suitable split points, so wait a bit before
				sending a HELP again. */
			if (x == -1)
			{
				/* We've been waiting for a while, so reset the counter and
					send out some more HELPs. */
				if (last_help_time > 2000 || help_count == 0)
					last_help_time = -1;
				else
					goto end;
			}
		}
		/* last_help_time is -1, meaning the other processors don't know we're
			idle. Send them HELP messages so we can get their tree info. */
		if (last_help_time == -1)
		{
			/* Make sure all the old search states are cleaned up. */
			for (process = 0; process < zct->process_count; process++)
				get_tree_block(process, board.id)->used = FALSE;
			/* Inform other processors we are idle. */
			help_count = 0;
			for (process = 0; process < zct->process_count; process++)
			{
				if (process != board.id && !smp_block[process].idle)
				{
					smp_message(process, SMP_HELP, board.id);
					help_count++;
				}
			}
			/* We were able to send a message, so start the wait counter. */
			if (help_count != 0)
				last_help_time = 0;
			goto end;
		}

		/* Here we have found a split point to attach to, so clean up a bit and
			then try to attach to it. */
found_split_point:
		last_help_time = -1;
		for (process = 0; process < zct->process_count; process++)
		{
			done[process] = 0;
			get_tree_block(process, board.id)->used = FALSE;
		}
		/* We found a split point, now try to attach to it. */
		temp = attach(x);
		/* Check if the split point is still good. */
		if (temp == NULL)
		{
			SMP_DEBUG(print("cpu %i received error on attach.\n", board.id));
			goto end;
		}
		/* Now, turn our search block to the split point and start searching! */
		sb = temp;
		SMP_DEBUG(print("cpu %i successfully attached. x=%i %E\n",
			board.id, x, sb));
		goto end;
case SEARCH_CHILD_RETURN:
		/* A child has returned to this point, which means that the main
			processor should return. */
		SMP_DEBUG(print("cpu %i flagging return, %E\n", board.id, sb + 1));
		ASSERT(board.game_entry == root_entry + 1);
		ASSERT(board.split_point == board.split_point_stack);
		ASSERT(board.split_ply == board.split_ply_stack);
		unmake_move();
		/* This lock should never be needed, as only this processor should be
			active, but just in case... */
		LOCK(smp_data->lock);
		smp_data->return_value = (sb + 1)->return_value;
		copy_pv(smp_data->return_pv, board.pv_stack[1]);
		smp_data->return_flag = TRUE;
		UNLOCK(smp_data->lock);
		sb++;
		sb->search_state = SEARCH_WAIT;
		set_idle();
		goto end;
#endif
/* We are stopping... */
case SEARCH_STOP:
		while (board.game_entry > root_entry)
			unmake_move();
		/* This is kind of hacky... */
		if (/*board.id == 0 &&*/ board.game_entry == root_entry)
			make_move((zct->next_root_move - 1)->move);

		return sb->alpha;
/* The master process has finished searching. Presumably all other processes
	are in SEARCH_WAIT. */
case SEARCH_RETURN:
#ifdef SMP
		set_active();
		/* This is kind of hacky... */
		if (board.id == 0 && board.game_entry == root_entry)
			make_move((zct->next_root_move - 1)->move);
#endif
		return (sb + 1)->return_value;
/* Do basic node processing: time, input, and messages in SMP. */
end:
#ifdef SMP
		/* Handle any asynchronous messages we have. */
		if (smp_block[board.id].message_count)
			handle_messages(&sb);
		/* See if someone is telling us to split. */
		if (smp_block[board.id].input == SMP_SPLIT)
		{
			if (sb->search_state == SEARCH_CHILD_RETURN ||
				sb->search_state == SEARCH_WAIT)
			{
				smp_block[board.id].input = 0;
				smp_block[board.id].data = -1;
				smp_block[board.id].output = SMP_DONE;
			}
			else
			{
				/* Determine the ply and the search_block id. */
				r = split(sb, smp_block[board.id].data % MAX_PLY,
					smp_block[board.id].data / MAX_PLY);
				smp_block[board.id].input = 0;
				smp_block[board.id].data = r;
				smp_block[board.id].output = SMP_DONE;
			}
		}
		else if (smp_block[board.id].input == SMP_PARK)
		{
			last_help_time = -1;
			smp_block[board.id].input = 0;
			smp_block[board.id].data = 0;
			smp_block[board.id].output = SMP_DONE;
			stop(sb);
			return 0;
		}
		/* If another processor has backed up to the root, they set a flag
			because only the main processor can return to search_root(). */
		if (board.id == 0 && smp_data->return_flag)
		{
			ASSERT(board.split_point == board.split_point_stack);
			ASSERT(smp_data->split_count == 0);
			SMP_DEBUG(print("cpu 0 returning...\n"));
			/* Back up the game board. */
			while (board.game_entry > root_entry + 1)
				unmake_move();
#ifdef SMP
			/* This is kind of hacky... */
			if (board.id == 0 && board.game_entry == root_entry)
				make_move((zct->next_root_move - 1)->move);
#endif
			smp_data->return_flag = FALSE;
			set_active();
			/* Return the score and PV. */
			copy_pv(board.pv_stack[1], smp_data->return_pv);
			return smp_data->return_value;
		}
#endif

		/* Do the standard periodic check for time or input. */
		if (
#ifdef SMP
			board.id == 0 &&
#endif
			zct->nodes_until_check-- <= 0)
		{
			zct->nodes_until_check = zct->nodes_per_check;
			if (stop_search())
			{
#ifdef SMP
				/* Flag the other processors down. */
				for (r = 1; r < zct->process_count; r++)
					smp_tell(r, SMP_PARK, 0);
				/* Unsplit all of our split points, after waiting for the
					child processors to detach. */
				while (*board.split_point != NULL)
				{
					while ((*board.split_point)->child_count > 1)
						;
					unsplit(sb, (*board.split_point)->id);
				}
#endif
				/* We're good to go. Stop the search. */
				sb->search_state = SEARCH_STOP;
				goto end;
			}
		}
	} /* <-- This brace is from the switch, which isn't indented. */
	}
}
